// Security System
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#if !defined ( __SECURITY_SYSTEM_H_INCLUDED__)
#define __SECURITY_SYSTEM_H_INCLUDED__

#include <time.h>

#if defined(_MSC_VER )
  // It's OK that the compiler can't automatically generate default constructors or
  // assignment operators for classes with const data members
  // We're shooting for immutability here
  #pragma warning ( disable : 4702 )  // unreachable code in <list>
  #if defined(SECURITY_SYSTEM_EXPORTS)
    #define SECURITY_SYSTEM_DLL _declspec(dllexport)
  #else //  defined ( SECURITY_SYSTEM_EXPORTS )
    #define SECURITY_SYSTEM_DLL _declspec(dllimport)
  #endif  // defined ( SECURITY_SYSTEM_EXPORTS )
#else
   #define SECURITY_SYSTEM_DLL
#endif

#include <TLV.h>
#include <list>
#include <vector>
#include <APDU_Packet.h>
#include <macros.h>

//
// Access Control List Tags
//
// Security Attribute Coding //
#define ACR_REFERENCE_FORM                      0x8B
#define ACR_COMPACT_FORM                        0x8C
#define ACR_EXPANDED_FORM                       0xAB
#define ACR_PROPRIETARY                         0x86

#define ACR_AM_DO                               0x80
#define ACR_AM_DO_CLA                           0x88
#define ACR_AM_DO_INS                           0x84
#define ACR_AM_DO_P1                            0x82
#define ACR_AM_DO_P2                            0x81
#define ACR_AM_ANY                              0xFF

//
// Access Modes for EFs
//
//    READ            - ReadBinary, ReadRecord, SearchBinary, SearchRecord
//    UPDATE          - UpdateBinary, UpdateRecord, EraseBinary
//    WRITE           - Write Binary, WriteRecord, AppendRecord, Increase
//    DEACTIVATE      - DeactivateFile
//    ACTIVATE        - ActivateFile
//    TERMINATE       - TerminateEF
//    DELETE          - DeleteFile
//
#define ACR_AM_FILE_ANY                         0xFF
#define ACR_AM_FILE_EXECUTE                     0x90
#define ACR_AM_FILE_DELETE                      0x40
#define ACR_AM_FILE_TERMINATE                   0x20
#define ACR_AM_FILE_REHABILITATE                0x10
#define ACR_AM_FILE_ACTIVATE                    0x10
#define ACR_AM_FILE_DEACTIVATE                  0x08
#define ACR_AM_FILE_INVALIDATE                  0x08
#define ACR_AM_FILE_APPEND                      0x04
#define ACR_AM_FILE_WRITE                       0x04
#define ACR_AM_FILE_INCREASE                    0x04
#define ACR_AM_FILE_UPDATE                      0x02
#define ACR_AM_FILE_SEARCH                      0x01
#define ACR_AM_FILE_READ                        0x01
// combinations of the above
#define ACR_AM_FILE_MODIFY                      ( ACR_AM_FILE_INCREASE | ACR_AM_FILE_UPDATE )

//
// Access Modes for DFs
//
//    DELETE          - DeleteFile (child)
//    CREATE_EF       - CreateFile - EF creation
//    CREATE_DF       - CreateFile - DF creation
//    DEACTIVATE      - DeactivateFile
//    ACTIVATE        - ActivateFile
//    TERMINATE       - TerminateCard, Terminate (directory)
//    DELETE_SELF     - DeleteFile (self)
//
#define ACR_AM_DIR_ANY                          0xFF
#define ACR_AM_DIR_DELETE_SELF                  0x40
#define ACR_AM_DIR_TERMINATE_CARD_USAGE         0x20
#define ACR_AM_DIR_TERMINATE                    0x20
#define ACR_AM_DIR_ACTIVATE                     0x10
#define ACR_AM_DIR_DEACTIVATE                   0x08
#define ACR_AM_DIR_CREATE_DF                    0x04
#define ACR_AM_DIR_CREATE_EF                    0x02
#define ACR_AM_DIR_DELETE_CHILD                 0x01

// Access Modes for Either EFs or DFs
#define ACR_AM_ANY                              0xFF
#define ACR_AM_TERMINATE                        0x20
#define ACR_AM_ACTIVATE                         0x10
#define ACR_AM_DEACTIVATE                       0x08

//
// Access Modes for Data Objects
//
//    GET_DATA        - Get Data
//    PUT_DATA        - Put Data
//    MSE             - Manage Security Environment
//
#define ACR_AM_DO_ANY                           0xFF
#define ACR_AM_DO_MSE                           0x04
#define ACR_AM_DO_PUT_DATA                      0x02
#define ACR_AM_DO_GET_DATA                      0x01

// 
// Security Conditions 
//
#define ACR_SC_DO                               0xA4
#define ACR_SC_DO_OR                            0xA0
#define ACR_SC_DO_NOT                           0xA7
#define ACR_SC_DO_AND                           0xAF
#define ACR_SC_DO_ALWAYS                        0x90
#define ACR_SC_DO_NEVER                         0x97

#define ACR_SC_KEY_REFERENCE                    0x83

#define ACR_SC_USAGE_QUALIFIER                  0x95
#define ACR_UQ_DO_NOT_USE                       0x00
#define ACR_UQ_EXTERNAL_AUTHENTICATION          0x80
#define ACR_UQ_INTERNAL_AUTHENTICATION          0x40
#define ACR_UQ_USER_AUTHENTICATION              0x08
#define ACR_UQ_USE                              0x08

//
// Manage Security Environment operation code
//
#define MSE_SET_DST                             0x41

//
// Perform Security operation code
//
#define PSO_DIGITAL_SIGNATURE                   0x9E

#define PSO_DIGITAL_SIGNATURE_DATA              0x9A

//
// Control Reference Templates
//
#define CRT_AUTHENTICATION                      0xA4
#define CRT_KEY_AGREEMENT                       0xA6
#define CRT_HASH_CODE                           0xAA
#define CRT_CRYPTOGRAPHIC_CHECKSUM              0xB4
#define CRT_DIGITAL_SIGNATURE                   0xB6
#define CRT_CONFIDENTIALITY                     0xB8

#define CRT_AUTHENTICATION_CHK                  0xA5
#define CRT_HASH_CODE_CHK                       0xAB
#define CRT_CRYPTOGRAPHIC_CHECKSUM_CHK          0xB5
#define CRT_DIGITAL_SIGNATURE_CHK               0xB7
#define CRT_CONFIDENTIALITY_CHK                 0xB9

//
// Control Reference Data Objects
//
#define CRT_ALGORITHM                           0x80
#define CRT_FID                                 0x81
#define CRT_DF_NAME                             0x82

// Symmetric Symmetric_Key
#define CRT_KEYREF_DIRECT_USE                   0x83
#define CRT_KEYREF_DERIVED_USE                  0x84
// Asymmetric Symmetric_Key
#define CRT_KEYREF_PUBLIC_KEY                   0x83
#define CRT_KEYREF_PRIVATE_KEY                  0x84

#define CRT_NULL_BLOCK                          0x85
#define CRT_CHAINING_BLOCK                      0x86
#define CRT_INITIAL_VALUE_BLOCK                 0x87
#define CRT_PREVIOUS_CHALLENGE                  0x88
#define CRT_CRYPTOGRAM_DESCRIPTOR_BYTE          0x8E
#define CRT_CARD_HASH_CODE                      0x90
#define CRT_RANDOM_NUMBER                       0x91
#define CRT_TIME_STAMP                          0x92
#define CRT_COUNTER                             0x93
#define CRT_CHALLENGE                           0x94
#define CRT_USAGE_QUALIFIER                     0x95

// Usage Qualifiers
#define UQ_DO_NOT_USE                           0x00
#define UQ_EXTERNAL_AUTHENTICATION              0x80
#define UQ_INTERNAL_AUTHENTICATION              0x40
#define UQ_USER_AUTHENTICATION_KNOWLEDGE        0x08
#define UQ_USER_AUTHENTICATION_BIOMETRIC        0x04
#define UQ_USE                                  0x08

// Cryptogram Descriptor Byte
#define CDB_SECRET_KEY_FOR_INFORMATION          0x10
#define CDB_SECRET_KEY_FOR_KEYS                 0x20
#define CDB_SECRET_KEY                          0x20
#define CDB_PRIVATE_KEY                         0x30
#define CDB_PASSWORD                            0x40
#define CDB_PUBLIC_KEY                          0x80

//
//  RSA Data Objects
//
#define PK_TEMPLATE                             0x7F, 0x49
#define PK_RSA_MODULUS                          0x81
#define PK_RSA_PUBLIC_EXPONENT                  0x82
#define PK_RSA_PRIVATE_EXPONENT                 0x83
#define PK_RSA_FACTOR_P                         0x84
#define PK_RSA_FACTOR_Q                         0x85
#define PK_RSA_MONTGOMERY_N0                    0x86
#define PK_RSA_EXPONENT_P                       0x87
#define PK_RSA_EXPONENT_Q                       0x88
#define PK_RSA_INV_Q_MOD_P                      0x89
#define PK_RSA_MONTGOMERY_P0                    0x8A
#define PK_RSA_MONTGOMERY_Q0                    0x8B

#define PK_ECC_POINT_X                          0x81
#define PK_ECC_POINT_Y                          0x82

#define MANDATORY_PIN_LENGTH                    0x08

#define SPECIFIC_REFERENCE_DATA                 0x80

// Table 13 - Key References
#define CARDHOLDER_PIV_CARD_APPLICATION_PIN             0x80    // 8-Byte PIN
#define CARDHOLDER_PIV_CARD_APPLICATION_PUK             0x81    // 8-Byte PIN Unblocking Key
#define PIV_CARD_APPLICATION_VALIDATION                 0x9A    // 1024-Bit RSA
#define PIV_CARD_APPLICATION_ADMINISTRATION             0x9B    // 192-Byte Triple DES ECB
#define PIV_CARD_APPLICATION_DIGITAL_SIGNATURE          0x9C    // 1024-Bit RSA
#define PIV_CARD_APPLICATION_KEY_MANAGEMENT             0x9D    // 1024-Bit RSA
#define PIV_CARD_AUTHENTICATION_KEY                     0x9E    // Unspecified

static bool interface_is_contactless = 0 ;

// Status_Reference is a wrapper on Octet that exhibits a value of one less
// It is also the home of the central key info related to key numbers
class SECURITY_SYSTEM_DLL Status_Reference
  {
  public:
    static
    void validate_status_reference (Status_Reference & status_number ,
                                    const Memory::size_t max_status_number ,
                                    Status_Reference the_default_status_reference )
      {
        if ( 0 == status_number . the_status_number )
          status_number = the_default_status_reference ;

        if ( ! ( 1 <= status_number . the_status_number 
               && status_number . the_status_number <= max_status_number ) )
          throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
      } ;
    
    Status_Reference ( Octet status_number ) 
      : the_status_number ( status_number )
      { } ;

    Octet as_index ( ) const
        { return the_status_number - 1  ; } ;

    bool operator == ( const Status_Reference & other) const
    {
        return the_status_number == other . the_status_number ;
    }

    bool operator != ( const Status_Reference & other) const
    {
        return the_status_number != other . the_status_number ;
    }

    private:
        Octet the_status_number ;
  } ;

// Security_Status is a wrapper for a vector of bits
class SECURITY_SYSTEM_DLL Security_Status: public Object
  {
    static const int number_of_bits_per_octet = 8 ;

  public:
    Security_Status ( const Memory::size_t number_of_status_references )
      : the_number_of_status_references ( number_of_status_references ) ,
        the_number_of_octets_in_bit_storage 
          ( ( the_number_of_status_references + number_of_bits_per_octet - 1 )
             / number_of_bits_per_octet ) ,
             the_bit_storage_pointer 
               ( Octets::Pointer::object_pointer_to_indexed_pointer 
                  ( new Octets (  the_number_of_octets_in_bit_storage ) ) )
      { reset_security_status ( ) ; } ;

    void authenticate ( const Status_Reference & status_reference ) 
      { 
        Memory::offset_t octet_index =
          status_reference . as_index ( ) / number_of_bits_per_octet ;
        Memory::offset_t bit_offset =
          status_reference . as_index ( ) % number_of_bits_per_octet ;
        the_bit_storage_pointer -> put ( octet_index , 
                                ( the_bit_storage_pointer -> get ( octet_index ) 
                                    | ( 1 << bit_offset ) ) ) ;
      } ;

    void deauthenticate ( const Status_Reference status_reference )
      { 
        Memory::offset_t octet_index =
          status_reference . as_index ( ) / number_of_bits_per_octet ;
        Memory::offset_t bit_offset =
          status_reference . as_index ( ) % number_of_bits_per_octet ;
        the_bit_storage_pointer -> put ( octet_index , 
                                ( the_bit_storage_pointer -> get ( octet_index ) 
                                    & ~( 1 << bit_offset ) ) ) ;
      } ;

    bool authenticated ( const Status_Reference status_reference ) const
      { 
        Memory::offset_t octet_index =
          status_reference . as_index ( ) / number_of_bits_per_octet ;
        Memory::offset_t bit_offset =
          status_reference . as_index ( ) % number_of_bits_per_octet ;
        return 0 != ( the_bit_storage_pointer -> get ( octet_index ) 
                        & ( 1 << bit_offset ) ) ;
      } ;

    void reset_security_status ( ) 
      { the_bit_storage_pointer -> set( 0 , the_number_of_octets_in_bit_storage ) ; } ;

  private:
    Memory::size_t the_number_of_status_references ;
    Memory::size_t the_number_of_octets_in_bit_storage ;
    Octets::Pointer the_bit_storage_pointer ;
  };

// Key_Reference is a wrapper on Octet that exhibits a value of one less
// It is also the home of the central key info related to key numbers
class SECURITY_SYSTEM_DLL Key_Reference: public Status_Reference
  {
  public:
    static const Memory::size_t number_of_key_references = 0x40 ;
    static
    void validate_key_reference (Key_Reference & status_number ,
                                 Key_Reference the_default_status_reference = 0 )
      {
        Status_Reference::validate_status_reference (status_number ,
                                                     number_of_key_references ,
                                                     the_default_status_reference ) ;
      } ;

    // Check for a valid PIV Card Application key reference and make the key reference namespace 
    // contiguous by shifting application-specific ("local") key references down
    Key_Reference ( Octet key_reference ) 
        : Status_Reference ( ( key_reference & SPECIFIC_REFERENCE_DATA ) ? 
                                             ( ( key_reference & 0x01F ) | 0x20 ) : key_reference ) 
      {  
            if(( key_reference & SPECIFIC_REFERENCE_DATA ) &&
               !(key_reference == CARDHOLDER_PIV_CARD_APPLICATION_PIN       ||
                 key_reference == CARDHOLDER_PIV_CARD_APPLICATION_PUK       ||
                 key_reference == PIV_CARD_APPLICATION_VALIDATION           ||
                 key_reference == PIV_CARD_APPLICATION_ADMINISTRATION       ||
                 key_reference == PIV_CARD_APPLICATION_DIGITAL_SIGNATURE    ||
                 key_reference == PIV_CARD_APPLICATION_KEY_MANAGEMENT       ||
                 key_reference == PIV_CARD_AUTHENTICATION_KEY ) )
            throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
      } ;
  } ;

// PIN_Store is a wrapper for 48 Octet_String pointers
typedef Octet_String PIN ;
typedef Octet_String_Pointer PIN_Pointer ;
typedef Octet PIN_Retries ;
struct PIN_Entry : public Object
  {
    PIN_Entry ( const PIN_Pointer & PIN ,
                const PIN_Pointer & PUK ,
                PIN_Retries remaining_number ,
                PIN_Retries default_number ,
                PIN_Retries unblock_number )
    : PIN_pointer ( PIN ) ,
      PUK_pointer ( PUK ) ,
      retries_remaining ( remaining_number ) ,
      default_retries ( default_number ) ,
      unblocks_remaining ( unblock_number ) 
      { } ;
      
    PIN_Pointer PIN_pointer ;
    PIN_Pointer PUK_pointer ;
    PIN_Retries retries_remaining ;
    PIN_Retries default_retries ;
    PIN_Retries unblocks_remaining ;
  } ;
typedef indexed_pointer < PIN_Entry > PIN_Entry_Pointer ;

class SECURITY_SYSTEM_DLL Algorithm_Identifier
  {
  public :
    static const Octet Default                      = 0x00 ; // The default PIV algorithm is Triple_DES_ECB_128
    static const Octet Triple_DES_ECB_128           = 0x01 ;
    static const Octet Triple_DES_CBC_128           = 0x02 ;
    static const Octet Triple_DES_ECB_192           = 0x03 ;
    static const Octet Triple_DES_CBC_192           = 0x04 ;
    static const Octet RSA_3072                     = 0x05 ;
    static const Octet RSA_1024                     = 0x06 ;
    static const Octet RSA_2048                     = 0x07 ;
    static const Octet AES_ECB_128                  = 0x08 ;
    static const Octet AES_CBC_128                  = 0x09 ;
    static const Octet AES_ECB_192                  = 0x0A ;
    static const Octet AES_CBC_192                  = 0x0B ;
    static const Octet AES_ECB_256                  = 0x0C ;
    static const Octet AES_CBC_256                  = 0x0D ;
    static const Octet ECC_P_224                    = 0x0E ;
    static const Octet ECC_K_233                    = 0x0F ;
    static const Octet ECC_B_233                    = 0x10 ;
    static const Octet ECC_P_256                    = 0x11 ;
    static const Octet ECC_K_283                    = 0x12 ;
    static const Octet ECC_B_283                    = 0x13 ;
    static const Octet Last_Algorithm_Identifier    = 0x13;

    Algorithm_Identifier ( Octet identifier )
        : the_identifier ( identifier == Default ?  Triple_DES_ECB_128 : identifier )
      { 
        if ( ! (    this -> is_DES_algorithm ( )
                 || this -> is_RSA_algorithm ( )
                 || this -> is_AES_algorithm ( )
                 || this -> is_ECC_algorithm ( ) ) )
            throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;
      } ;

    operator Octet ( ) { return the_identifier ; } ;

    bool is_DES_algorithm ( )
    {
        return    Triple_DES_ECB_128 == the_identifier 
               || Triple_DES_CBC_128 == the_identifier
               || Triple_DES_ECB_192 == the_identifier
               || Triple_DES_CBC_192 == the_identifier ; 
    }

    bool is_RSA_algorithm ( )
    {
        return    RSA_1024 == the_identifier 
               || RSA_2048 == the_identifier 
               || RSA_3072 == the_identifier ; 
    }

    bool is_AES_algorithm ( )
    {
        return    AES_ECB_128 == the_identifier 
               || AES_CBC_128 == the_identifier
               || AES_ECB_192 == the_identifier
               || AES_CBC_192 == the_identifier
               || AES_ECB_256 == the_identifier
               || AES_CBC_256 == the_identifier ; 
    }

    bool is_ECC_algorithm ( )
    {
        return    ECC_P_224 == the_identifier 
               || ECC_K_233 == the_identifier
               || ECC_B_233 == the_identifier
               || ECC_P_256 == the_identifier
               || ECC_K_283 == the_identifier
               || ECC_B_283 == the_identifier ; 
    }

    Octet the_identifier ;
  } ;

typedef Octet_String Symmetric_Key ;
typedef Octet_String_Pointer Symmetric_Key_Pointer ;

struct Symmetric_Key_Entry : public Object
  {
    Symmetric_Key_Entry ( const Symmetric_Key_Pointer & key )
      : key_pointer ( key ) 
      { } ;
    Symmetric_Key_Pointer key_pointer ;
  } ;


typedef indexed_pointer < Symmetric_Key_Entry > Symmetric_Key_Entry_Pointer ;


// RSA_Key_Store is a wrapper for 48 RSA Keys
class SECURITY_SYSTEM_DLL RSA_Key : public Object
  {
  public:
    typedef indexed_pointer < RSA_Key > Pointer;
    
//  Three Octet_Strings storing the modulus, private key and exponent
    RSA_Key ( const Octet * m_octets  , size_t m_size  ,
              const Octet * d_octets  , size_t d_size  ,
              const Octet * e_octets  , size_t e_size  )
      : m  ( m_octets  , m_size  ) ,
        d  ( d_octets  , d_size  ) ,
        e  ( e_octets  , e_size  ) 
      { } ;

    Octet_String m ;
    Octet_String d ;
    Octet_String e ;
    Octet * m_data ( ) { return m . data ( ) ; } ;
    Octet * d_data ( ) { return d . data ( ) ; } ;
    Octet * e_data ( ) { return e . data ( ) ; } ;
  } ;

typedef RSA_Key::Pointer RSA_Key_Pointer ;

struct RSA_Key_Entry : public Object
  {
    RSA_Key_Entry ( const RSA_Key_Pointer & key )
      : key_pointer ( key ) 
      { } ;
    RSA_Key_Pointer key_pointer ;
  } ;


typedef indexed_pointer < RSA_Key_Entry > RSA_Key_Entry_Pointer ;

// ECC_Key_Store is a wrapper for 48 ECC Keys
class SECURITY_SYSTEM_DLL ECC_Key : public Object
  {
  public:
    typedef indexed_pointer < ECC_Key > Pointer;
    
    ECC_Key ( const Octet * p_octets  , size_t p_size  ,
              const Octet * x_octets  , size_t x_size  ,
              const Octet * y_octets  , size_t y_size  )
      : p  ( p_octets  , p_size  ) ,
        x  ( x_octets  , x_size  ) ,
        y  ( y_octets  , y_size  ) 
      { } ;

    Octet_String p ;
    Octet_String x ;
    Octet_String y ;

    Octet * p_data ( ) { return p . data ( ) ; } ;
    Octet * x_data ( ) { return x . data ( ) ; } ;
    Octet * y_data ( ) { return y . data ( ) ; } ;
  } ;

typedef ECC_Key::Pointer ECC_Key_Pointer ;

struct ECC_Key_Entry : public Object
  {
    ECC_Key_Entry ( const ECC_Key_Pointer & key )
      : key_pointer ( key ) 
      { } ;
    ECC_Key_Pointer key_pointer ;
  } ;


typedef indexed_pointer < ECC_Key_Entry > ECC_Key_Entry_Pointer ;


class SECURITY_SYSTEM_DLL Wrong_Key_Entry_Pointer_Type_Exception { } ;
class SECURITY_SYSTEM_DLL Key_Entry_Pointer : public indexed_pointer_base
  {
  public:

    typedef Octet Type;
    static const Type generic       = 0 ;
    static const Type PIN           = 1 ;
    static const Type symmetric_key = 2 ;
    static const Type RSA_key       = 3 ;
    static const Type ECC_key       = 4 ;

    Key_Entry_Pointer ( )
      : indexed_pointer_base ( ) ,
        the_type ( generic )
      { } ;

    Key_Entry_Pointer ( const PIN_Entry_Pointer & p )
      : the_type ( PIN )
      { add_reference_to_referenced_object ( p . object_pointer ( ) ) ; } ;

    Key_Entry_Pointer ( const Symmetric_Key_Entry_Pointer & p )
      : the_type ( symmetric_key )
      { add_reference_to_referenced_object ( p . object_pointer ( ) ) ; } ;

    Key_Entry_Pointer ( const RSA_Key_Entry_Pointer & p )
      : the_type ( RSA_key )
      { add_reference_to_referenced_object ( p . object_pointer ( ) ) ; } ;

    Key_Entry_Pointer ( const ECC_Key_Entry_Pointer & p )
      : the_type ( ECC_key )
      { add_reference_to_referenced_object ( p . object_pointer ( ) ) ; } ;

    operator PIN_Entry_Pointer ( )
      {
        if ( PIN == the_type )
          return 
            PIN_Entry_Pointer::object_pointer_to_indexed_pointer 
              ( reinterpret_cast < PIN_Entry * > ( object_pointer ( ) ) ) ;

        throw Wrong_Key_Entry_Pointer_Type_Exception ( ) ;
      } ;
  
    operator Symmetric_Key_Entry_Pointer ( )
      {
        if ( symmetric_key == the_type )
          return 
            Symmetric_Key_Entry_Pointer::object_pointer_to_indexed_pointer 
              ( reinterpret_cast < Symmetric_Key_Entry * > ( object_pointer ( ) ) ) ;

        throw Wrong_Key_Entry_Pointer_Type_Exception ( ) ;
      } ;
  
    operator RSA_Key_Entry_Pointer ( )
      {
        if ( RSA_key == the_type )
          return 
            RSA_Key_Entry_Pointer::object_pointer_to_indexed_pointer 
              ( reinterpret_cast < RSA_Key_Entry * > ( object_pointer ( ) ) ) ;

        throw Wrong_Key_Entry_Pointer_Type_Exception ( ) ;
      } ;
  
    operator ECC_Key_Entry_Pointer ( )
      {
        if ( ECC_key == the_type )
          return 
            ECC_Key_Entry_Pointer::object_pointer_to_indexed_pointer 
              ( reinterpret_cast < ECC_Key_Entry * > ( object_pointer ( ) ) ) ;

        throw Wrong_Key_Entry_Pointer_Type_Exception ( ) ;
      } ;
  
    Type type ( ) const { return the_type ; } ;

    Type the_type ;

  } ;

class SECURITY_SYSTEM_DLL Key_Entry_Pointers : public storage < Key_Entry_Pointer > { } ;

class SECURITY_SYSTEM_DLL Key_Store : public Object
  {
  public:
    Key_Store ( )
      : the_key_entries_pointers_pointer 
        ( Key_Entry_Pointers::alloc ( Key_Reference::number_of_key_references ) )
      { } ;

    Key_Entry_Pointer get_key_entry_pointer ( const Key_Reference key_reference ) const
      { return the_key_entries_pointers_pointer -> get ( key_reference . as_index ( ) ) ; } ;

    Key_Entry_Pointer put_key_entry_pointer ( const Key_Reference key_reference , 
                                              const Key_Entry_Pointer & Key_entry_pointer )
      { 
        the_key_entries_pointers_pointer -> put ( key_reference . as_index ( ) , Key_entry_pointer ) ;
        return Key_entry_pointer  ;
      } ;

    private:
       Key_Entry_Pointers::Pointer the_key_entries_pointers_pointer ;      
  } ;


class SECURITY_SYSTEM_DLL Key_Reference_Information : public Primitive_BER_TLV
  {
  public :
    Key_Reference_Information ( const Octet * octets, size_t size )
      : Primitive_BER_TLV ( Octet_String ( octets , size ) )
      {
        if ( BER_TLV_Tag ( CRT_KEYREF_PRIVATE_KEY ) != the_tag )
          throw RAPDU_Packet::SW_WRONG_PARAMETERS_IN_DATA_FIELD ;

        if ( 1 != the_length . value ( ) )
          throw RAPDU_Packet::SW_WRONG_PARAMETERS_IN_DATA_FIELD ;
      } ;

    Key_Reference key_reference ( )
      { return Key_Reference ( the_values . get_octet ( 0 ) ) ; } ;
  } ;



// Security_Environment needs more semantics
class SECURITY_SYSTEM_DLL Security_Environment : public Object
  {
  public:
    Security_Environment ( Key_Reference key_reference = 0 )
      : the_key_reference ( key_reference ) ,
        the_algorithm_identifier ( Algorithm_Identifier::RSA_1024 ) 
      { } ;

    Key_Reference the_key_reference ;

    Algorithm_Identifier the_algorithm_identifier ;

  };

/* ISO/IEC 8825-1 multi-byte Octet_String */
class SECURITY_SYSTEM_DLL Access_Mode_Data_Object 
  {
  public:

    // TODO: handle Command header descriptions?
    Access_Mode_Data_Object ( const Primitive_BER_TLV_Pointer data_object_pointer )
    { 
      const Primitive_BER_TLV & the_TLV = * data_object_pointer ;
      if ( ! is_access_mode_tag ( the_TLV . tag ( ) ) )
        {
          debug_message("Access_Mode_Data_Object:tag is not access mode tag in the_TLV=%s\n",
                        the_TLV . spell ( ) . c_str ( )) ;
          throw RAPDU_Packet::SW_INCOMPATIBLE_FILE_STRUCTURE ;
        }

      // There should just be one access mode byte for now
      const BER_TLV_Length the_length = the_TLV . length ( ) ;
      const Memory::size_t the_length_value = the_length . value ( ) ;
      // There should just be one access mode byte for now
      if ( 1 != the_length_value )
        throw RAPDU_Packet::SW_INCOMPATIBLE_FILE_STRUCTURE ;

      the_access_mode = data_object_pointer -> values ( ) . get_octet ( 0 ) ;

    } ;

    static bool is_access_mode_tag ( const BER_TLV_Tag & tag )
      { return BER_TLV_Tag ( ACR_AM_DO ) == tag ; } ;

    bool mode_matches ( Octet access_mode ) const
      { return 0 != ( access_mode & the_access_mode ) ; } ;

    Octet the_access_mode ;

  } ;



class SECURITY_SYSTEM_DLL Primitive_Security_Condition_Data_Object ;

class SECURITY_SYSTEM_DLL Constructed_Security_Condition_Data_Object ;

class SECURITY_SYSTEM_DLL Security_Condition_Data_Object;

class SECURITY_SYSTEM_DLL Wrong_Security_Condition_Data_Object_Pointer_Type_Exception { } ;

class SECURITY_SYSTEM_DLL Security_Condition_Data_Object_Pointer : public indexed_pointer_base
  {
  public:
    // This would be an enum were space not at a premium
    typedef Octet Type ;

    static const Type generic      = 0 ;
    static const Type primitive    = 1 ;
    static const Type constructed  = 2 ;

    Security_Condition_Data_Object_Pointer ( )
      : the_type ( generic )
      { } ;

    Security_Condition_Data_Object_Pointer ( const Security_Condition_Data_Object_Pointer & _Other )
      {
        attach ( Memory::index_to_pointer ( _Other . the_object_table_index ) ) ;
        the_type = _Other . the_type ;
      } ;

    Security_Condition_Data_Object_Pointer ( Primitive_Security_Condition_Data_Object * primitive_Security_Condition_Data_Object_pointer )
      {
        attach ( primitive_Security_Condition_Data_Object_pointer ) ;
        the_type = primitive ;
      } ;

    Security_Condition_Data_Object_Pointer ( Constructed_Security_Condition_Data_Object * constructed_Security_Condition_Data_Object_pointer )
      {
        attach ( constructed_Security_Condition_Data_Object_pointer ) ;
        the_type = constructed ;
      } ;

    ~Security_Condition_Data_Object_Pointer ( ) 
      { detach ( object_pointer ( ) ) ; } ;

    static Security_Condition_Data_Object_Pointer null ( ) { return Security_Condition_Data_Object_Pointer ( ) ; } ;

    Security_Condition_Data_Object_Pointer & operator = ( const Security_Condition_Data_Object_Pointer & _Other ) 
      {
        detach ( Memory::index_to_pointer (          the_object_table_index ) ) ;
        attach ( Memory::index_to_pointer ( _Other . the_object_table_index ) ) ;
        return * this ;
      } ;

    const Security_Condition_Data_Object * operator -> ( ) const 
      throw ( Wrong_Security_Condition_Data_Object_Pointer_Type_Exception ) ;

    Security_Condition_Data_Object * operator -> ( )
      throw ( Wrong_Security_Condition_Data_Object_Pointer_Type_Exception ) ;

    Type get_type ( ) const { return the_type ; } ;

  protected:
    void attach ( void * the_object_pointer ) ;

    void detach ( void * the_object_pointer ) ;
   
    Type the_type ;
  } ;

class SECURITY_SYSTEM_DLL Security_Condition_Data_Object : public BER_TLV
{
public:

   // Create a Constructed_Security_Condition_Data_Object or a Primitive_Security_Condition_Data_Object
   // depending on the outermost tag
   static Security_Condition_Data_Object_Pointer create 
     ( const BER_TLV_Pointer & data_object_pointer );

protected:
  // The constructors are protected so that only subclasses can be instantiated
   Security_Condition_Data_Object ( const BER_TLV & data_object )
      : BER_TLV ( data_object )
   { };
} ;


class SECURITY_SYSTEM_DLL Primitive_Security_Condition_Data_Object_Pointer 
  : public indexed_pointer < Primitive_Security_Condition_Data_Object > 
  {
  public:
    Primitive_Security_Condition_Data_Object_Pointer ( ) { } ;
    Primitive_Security_Condition_Data_Object_Pointer ( const BER_TLV_Pointer & data_object_pointer )
      {
        if ( BER_TLV_Pointer::primitive == data_object_pointer . get_type ( ) ) 
          attach ( data_object_pointer . object_pointer ( ) );
        else
          throw Wrong_Security_Condition_Data_Object_Pointer_Type_Exception ( ) ;
      } ;
    Primitive_Security_Condition_Data_Object_Pointer ( const Security_Condition_Data_Object_Pointer & data_object_pointer )
      {
        if ( Security_Condition_Data_Object_Pointer::primitive == data_object_pointer . get_type ( ) ) 
          attach ( data_object_pointer . object_pointer ( ) );
        else
          throw Wrong_Security_Condition_Data_Object_Pointer_Type_Exception ( ) ;
      } ;
  } ;


class SECURITY_SYSTEM_DLL Security_System ;
class SECURITY_SYSTEM_DLL Primitive_Security_Condition_Data_Object : public Primitive_BER_TLV
{
public:
  friend class Primitive_Security_Condition_Data_Object_Pointer ;
  Primitive_Security_Condition_Data_Object ( const Primitive_BER_TLV & data_object )
     : Primitive_BER_TLV ( data_object )
  { } ;

  bool evaluate ( const Security_System & ) const
    {
      if ( the_tag == BER_TLV_Tag ( ACR_SC_DO_ALWAYS ) )
        return true ;
      else if ( the_tag == BER_TLV_Tag ( ACR_SC_DO_NEVER ) )
        return false ;
      else
        return false ;
    } ;
} ;


class SECURITY_SYSTEM_DLL Constructed_Security_Condition_Data_Object ;

class SECURITY_SYSTEM_DLL Constructed_Security_Condition_Data_Object_Pointer 
  : public indexed_pointer < Constructed_Security_Condition_Data_Object > 
  {
  public:
    Constructed_Security_Condition_Data_Object_Pointer ( const BER_TLV_Pointer & data_object_pointer )
      {
        if ( BER_TLV_Pointer::constructed == data_object_pointer . get_type ( ) ) 
          attach ( data_object_pointer . object_pointer ( ) );
        else
          throw Wrong_Security_Condition_Data_Object_Pointer_Type_Exception ( ) ;
      } ;
    Constructed_Security_Condition_Data_Object_Pointer ( const Security_Condition_Data_Object_Pointer & data_object_pointer )
      {
        if ( Security_Condition_Data_Object_Pointer::constructed == data_object_pointer . get_type ( ) ) 
          attach ( data_object_pointer . object_pointer ( ) );
        else
          throw Wrong_Security_Condition_Data_Object_Pointer_Type_Exception ( ) ;
      } ;
  } ;


class SECURITY_SYSTEM_DLL Security_Condition_Data_Object_Sequence 
  : public list < Security_Condition_Data_Object_Pointer , 
                  persistent_allocator < Security_Condition_Data_Object_Pointer > >
  {
  public:
    Security_Condition_Data_Object_Sequence ( BER_TLV_Sequence BER_TLV_sequence )
      {
        for ( BER_TLV_Sequence::iterator p = BER_TLV_sequence . begin ( ) ;
              p != BER_TLV_sequence . end ( ) ;
              p ++ )
          push_back ( Security_Condition_Data_Object::create ( * p ) ) ;
      } ;

    Security_Condition_Data_Object_Sequence ( ) { } ;

    bool evaluate_status  ( const Security_System & security_system ) const ;
    bool evaluate_AND     ( const Security_System & security_system ) const ;
    bool evaluate_OR      ( const Security_System & security_system ) const ;
    bool evaluate_NOT     ( const Security_System & security_system ) const ;
  } ;

class SECURITY_SYSTEM_DLL Constructed_Security_Condition_Data_Object 
  : public Constructed_BER_TLV
  {
  public:
    friend class Constructed_Security_Condition_Data_Object_Pointer ;

    Constructed_Security_Condition_Data_Object ( const Constructed_BER_TLV & data_object )
      : Constructed_BER_TLV ( data_object )
   { } ;

    bool evaluate ( const Security_System & security_system ) const
      {
        Security_Condition_Data_Object_Sequence operands ( values ( ) ) ;
        if ( the_tag == BER_TLV_Tag ( ACR_SC_DO ) )
          return operands . evaluate_status ( security_system ) ;
        else  if ( the_tag == BER_TLV_Tag ( ACR_SC_DO_AND ) )
          return operands . evaluate_AND ( security_system ) ;
        else  if ( the_tag == BER_TLV_Tag ( ACR_SC_DO_OR ) )
          return operands . evaluate_OR ( security_system ) ;
        else  if ( the_tag == BER_TLV_Tag ( ACR_SC_DO_NOT ) )
          return operands . evaluate_NOT ( security_system ) ;
        else
          return false ;
    } ;
  } ;

class SECURITY_SYSTEM_DLL Access_Rule : public Object
  {
  public:
    Access_Rule ( BER_TLV_Sequence::iterator & p ,
                  BER_TLV_Sequence::iterator end )
      : the_access_mode_data_object ( * p )
      {
        while ( ++ p != end 
                  && ! Access_Mode_Data_Object::is_access_mode_tag ( ( * p ) -> tag ( ) ) )
          {
            Security_Condition_Data_Object_Pointer scdop =
              Security_Condition_Data_Object::create ( * p ) ;

            the_security_conditions . push_back ( scdop ) ;
          }
      }

    Access_Mode_Data_Object the_access_mode_data_object ;

    bool mode_matches ( Octet access_mode ) const
      { return the_access_mode_data_object . mode_matches ( access_mode ) ; } ; 

    bool evaluate ( const Security_System & security_system ) const
      {
        // The security conditions are implicitly ANDed
        return 
          the_security_conditions . evaluate_AND ( security_system ) ;
      } ;

    Security_Condition_Data_Object_Sequence the_security_conditions ;
  } ;

typedef indexed_pointer < Access_Rule > Access_Rule_Pointer ;

typedef  list < Access_Rule_Pointer , 
                persistent_allocator < Access_Rule_Pointer > > 
  Access_Rule_Pointer_Sequence ;


class SECURITY_SYSTEM_DLL Security_Attribute : public Constructed_BER_TLV
  {
  public:

    Security_Attribute ( const Constructed_BER_TLV & the_ACR )
      : Constructed_BER_TLV ( the_ACR )
      { 
        debug_message("Security_Attribute :parsing the_ACR=%s\n",
                      the_ACR . spell ( ) . c_str ( )) ;
        if ( tag ( ) != BER_TLV_Tag ( ACR_EXPANDED_FORM ) )
          throw RAPDU_Packet::SW_INCOMPATIBLE_FILE_STRUCTURE ;
        // Parse a sequence of Access Rules
        for ( BER_TLV_Sequence::iterator p = the_values . begin ( ) ;
              p != the_values . end ( ) ;
              // p advanced by Access_Rule constructor
            )
          the_access_rule_pointers . push_back 
            ( Access_Rule_Pointer::object_pointer_to_indexed_pointer 
                ( new Access_Rule ( p , the_values . end ( ) ) ) ) ;
      } ;

    bool authorized ( const Security_System & security_system , 
                      Octet access_mode ) const ;

    Access_Rule_Pointer_Sequence the_access_rule_pointers ; 
  };
typedef indexed_pointer < Security_Attribute > Security_Attribute_Pointer ;



// The Security System is the wrapper for security operations
// This version is designed to work with a single global security
// framework, although the components are designed to be resuable in
// a security architecture in which more than one key namespace is
// used, security status is maintained on a per-DF basis, etc.
class SECURITY_SYSTEM_DLL Security_System : public Object
  {
  public:
    Security_System ( bool entropy = false , 
                      Key_Reference default_PIN_number = 1  , 
                      Key_Reference default_symmetric_key_number = 10 , 
                      Key_Reference default_RSA_key_number = 11 )
      : entropy_is_enabled ( entropy ) ,
        the_default_PIN_number ( default_PIN_number ) ,
        the_default_symmetric_key_number ( default_symmetric_key_number ) ,
        the_default_RSA_key_number ( default_RSA_key_number ) ,
        the_security_status ( Key_Reference::number_of_key_references )
      { 
        if ( entropy )  
          srand ( static_cast < unsigned int > ( clock ( ) ) ) ;
      } ;

    static void set_interface ( bool contactless )
    {
        interface_is_contactless = contactless ;
    }

    static bool contactless_interface ( void )
    {
        return interface_is_contactless ;
    }

    void set_authentication ( Key_Reference the_key_reference  , bool authentication )
      {
        Key_Reference::validate_key_reference ( the_key_reference , 
                                                the_default_symmetric_key_number ) ;
        if ( authentication )
           the_security_status . authenticate ( the_key_reference ) ;
        else
           the_security_status . deauthenticate ( the_key_reference ) ;
      } ;

    void verify_pin ( Key_Reference the_key_reference , 
                      Octet_String the_presented_PIN ) 
      {
        Key_Reference::validate_key_reference ( the_key_reference , 
                                                the_default_PIN_number ) ;
  
        PIN_Entry_Pointer the_PIN_entry_pointer =
          the_key_store . get_key_entry_pointer ( the_key_reference );

        if ( PIN_Entry_Pointer ( ) == the_PIN_entry_pointer )
          throw RAPDU_Packet::SW_REFERENCED_DATA_NOT_FOUND ;

        PIN_Entry & the_PIN_entry = * the_PIN_entry_pointer ;

        Memory::size_t length = the_presented_PIN . length ( ) ;

        if ( MANDATORY_PIN_LENGTH != length )
          throw static_cast < RAPDU_Packet::Status_Word > 
                    ( RAPDU_Packet::SW_0_RETRIES_REMAINING 
                        + the_PIN_entry . retries_remaining ) ;
        else if ( 0 == the_PIN_entry . retries_remaining )
          throw RAPDU_Packet::SW_AUTHENTICATION_METHOD_BLOCKED ;
        else 
          {
            PIN & the_stored_PIN = * ( the_PIN_entry . PIN_pointer ) ; 
            if ( length == the_stored_PIN . length( ) 
                    && 0 == Octet_String::cmp( the_stored_PIN , 
                                              the_presented_PIN , 
                                              length ) )
            {
              the_PIN_entry . retries_remaining =
                the_PIN_entry . default_retries ;
              the_security_status . authenticate ( the_key_reference ) ;
            }
          else
            {
              the_security_status . deauthenticate ( the_key_reference ) ;
              throw static_cast < RAPDU_Packet::Status_Word > 
                        ( RAPDU_Packet::SW_0_RETRIES_REMAINING 
                            + ( -- the_PIN_entry . retries_remaining ) ) ;
            }
          }
      } ;

    void change_pin ( Key_Reference the_key_reference , 
                      Octet_String the_presented_PIN, 
                      Octet_String the_new_PIN ) 
      {
        verify_pin ( the_key_reference, the_presented_PIN ) ;
  
        PIN_Entry_Pointer the_PIN_entry_pointer =
          the_key_store . get_key_entry_pointer ( the_key_reference );

        PIN_Entry & the_PIN_entry = * the_PIN_entry_pointer ;

        Memory::size_t length = the_new_PIN . length ( ) ;

        if ( length != MANDATORY_PIN_LENGTH )
          throw RAPDU_Packet::SW_WRONG_LENGTH ;

        if ( PIN_Entry_Pointer ( ) == the_PIN_entry_pointer )
          throw RAPDU_Packet::SW_REFERENCED_DATA_NOT_FOUND ;

        PIN & the_stored_PIN = * ( the_PIN_entry . PIN_pointer ) ; 

        Octet_String::cpy( the_stored_PIN , 
                           the_new_PIN , 
                           MANDATORY_PIN_LENGTH ) ;

        the_PIN_entry . retries_remaining =
                the_PIN_entry . default_retries ;
    };

    void unblock_pin ( Key_Reference the_key_reference , 
                       Octet_String the_presented_PUK ,
                       Octet_String the_new_PIN ) 
      {
        Key_Reference::validate_key_reference ( the_key_reference , 
                                                the_default_PIN_number ) ;
  
        PIN_Entry_Pointer the_PIN_entry_pointer =
          the_key_store . get_key_entry_pointer ( the_key_reference );

        if ( PIN_Entry_Pointer ( ) == the_PIN_entry_pointer )
          throw RAPDU_Packet::SW_REFERENCED_DATA_NOT_FOUND ;

        PIN_Entry & the_PIN_entry = * the_PIN_entry_pointer ;

        Memory::size_t length = the_presented_PUK . length ( ) ;

        if ( MANDATORY_PIN_LENGTH != length )
          throw static_cast < RAPDU_Packet::Status_Word > 
                    ( RAPDU_Packet::SW_0_RETRIES_REMAINING 
                        + the_PIN_entry . unblocks_remaining ) ;
        else if ( 0 == the_PIN_entry . unblocks_remaining )
          throw RAPDU_Packet::SW_AUTHENTICATION_METHOD_BLOCKED ;
        else 
          {
            PIN & the_stored_PUK = * ( the_PIN_entry . PUK_pointer ) ; 
            if ( length == the_stored_PUK . length( ) 
                    && 0 == Octet_String::cmp( the_stored_PUK , 
                                              the_presented_PUK , 
                                              length ) )
            {
                the_PIN_entry . retries_remaining = the_PIN_entry . default_retries ;
//                the_security_status . authenticate ( the_key_reference ) ;
                PIN & the_stored_PIN = * ( the_PIN_entry . PIN_pointer ) ; 

                Octet_String::cpy( the_stored_PIN , the_new_PIN , MANDATORY_PIN_LENGTH ) ;
            }
          else
            {
              the_security_status . deauthenticate ( the_key_reference ) ;

              throw static_cast < RAPDU_Packet::Status_Word > 
                        ( RAPDU_Packet::SW_0_RETRIES_REMAINING 
                            + ( -- the_PIN_entry . unblocks_remaining ) ) ;
            }
          }
      }

    void generate_cryptogram ( Octet_String & cryptogram , Memory::size_t length )
      {
        if ( entropy_is_enabled )
          {
            for ( Memory::offset_t i = 0 ; i < length ; i ++ )
              cryptogram . put_octet ( i , static_cast < Octet > ( rand ( ) % 256 ) ) ;
          }
        else
          {
            Octet non_random_octet = 0 ;
            for ( Memory::offset_t i = 0 ; i < length ; i ++ )
              cryptogram . put_octet ( i , ++non_random_octet ) ;
          }
      }


    Symmetric_Key & validate_algorithm_and_symmetric_key ( Algorithm_Identifier algorithm_identifier ,
                                                           Key_Reference & key_number )
      {
        if ( Algorithm_Identifier::Last_Algorithm_Identifier < algorithm_identifier )
          throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;

        Key_Reference::validate_key_reference ( key_number , 
                                                the_default_symmetric_key_number ) ;
        
        Symmetric_Key_Entry_Pointer key_entry_pointer = 
          the_key_store . get_key_entry_pointer ( key_number ) ;

        if ( key_entry_pointer  == Symmetric_Key_Entry_Pointer::null ( ) )
          throw RAPDU_Packet::SW_CONDITIONS_OF_USE_NOT_SATISFIED ;
        
        return * ( key_entry_pointer -> key_pointer )  ;
      }


    RSA_Key & validate_algorithm_and_RSA_key ( Algorithm_Identifier algorithm_identifier ,
                                               Key_Reference & key_number )
      {
        if ( Algorithm_Identifier::Last_Algorithm_Identifier < algorithm_identifier )
          throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;

        Key_Reference::validate_key_reference ( key_number , 
                                                the_default_RSA_key_number ) ;

        RSA_Key_Entry_Pointer key_entry_pointer = 
          the_key_store . get_key_entry_pointer ( key_number ) ;

        if ( key_entry_pointer  == RSA_Key_Entry_Pointer::null ( ) )
          throw RAPDU_Packet::SW_CONDITIONS_OF_USE_NOT_SATISFIED ;
        
        return * ( key_entry_pointer -> key_pointer )  ;
      }

           
    ECC_Key & validate_algorithm_and_ECC_key ( Algorithm_Identifier algorithm_identifier ,
                                               Key_Reference & key_number )
      {
        if ( Algorithm_Identifier::Last_Algorithm_Identifier < algorithm_identifier )
          throw RAPDU_Packet::SW_INCORRECT_P1_P2 ;

        Key_Reference::validate_key_reference ( key_number ) ;
        
        ECC_Key_Entry_Pointer key_entry_pointer = 
          the_key_store . get_key_entry_pointer ( key_number ) ;

        if ( key_entry_pointer  == ECC_Key_Entry_Pointer::null ( ) )
          throw RAPDU_Packet::SW_CONDITIONS_OF_USE_NOT_SATISFIED ;
        
        return * ( key_entry_pointer -> key_pointer )  ;
      }


    void Triple_DES_ECB_128_encrypt ( Symmetric_Key & key , const Octet_String & plaintext , Octet_String & cyphertext ) ;

    void Triple_DES_ECB_128_decrypt ( Symmetric_Key & key , const Octet_String & cyphertext , Octet_String & plaintext ) ;


    void Triple_DES_CBC_128_encrypt ( Symmetric_Key & key , const Octet_String & plaintext , Octet_String & cyphertext ) ;

    void Triple_DES_CBC_128_decrypt ( Symmetric_Key & key , const Octet_String & cyphertext , Octet_String & plaintext ) ;


    void Triple_DES_ECB_192_encrypt ( Symmetric_Key & key , const Octet_String & plaintext , Octet_String & cyphertext ) ;

    void Triple_DES_ECB_192_decrypt ( Symmetric_Key & key , const Octet_String & cyphertext , Octet_String & plaintext ) ;


    void Triple_DES_CBC_192_encrypt ( Symmetric_Key & key , const Octet_String & plaintext , Octet_String & cyphertext ) ;

    void Triple_DES_CBC_192_decrypt ( Symmetric_Key & key , const Octet_String & cyphertext , Octet_String & plaintext ) ;


    void RSA_Test_encrypt ( RSA_Key & key , const Octet_String & plaintext , Octet_String & cyphertext ) ;

    void RSA_Test_decrypt ( RSA_Key & key , const Octet_String & cyphertext , Octet_String & plaintext ) ;

           
    void RSA_1024_encrypt ( RSA_Key & key , const Octet_String & plaintext , Octet_String & cyphertext ) ;

    void RSA_1024_decrypt ( RSA_Key & key , const Octet_String & cyphertext , Octet_String & plaintext ) ;

          
    void RSA_2048_encrypt ( RSA_Key & key , const Octet_String & plaintext , Octet_String & cyphertext ) ;

    void RSA_2048_decrypt ( RSA_Key & key , const Octet_String & cyphertext , Octet_String & plaintext ) ;


    void RSA_3072_encrypt ( RSA_Key & key , const Octet_String & plaintext , Octet_String & cyphertext ) ;

    void RSA_3072_decrypt ( RSA_Key & key , const Octet_String & cyphertext , Octet_String & plaintext ) ;
     

    void ECC_P_224_encrypt ( ECC_Key & key , const Octet_String & plaintext , Octet_String & cyphertext ) ;

    void ECC_P_224_decrypt ( ECC_Key & key , const Octet_String & cyphertext , Octet_String & plaintext ) ;

          
    void ECC_K_233_encrypt ( ECC_Key & key , const Octet_String & plaintext , Octet_String & cyphertext ) ;

    void ECC_K_233_decrypt ( ECC_Key & key , const Octet_String & cyphertext , Octet_String & plaintext ) ;


    void ECC_B_233_encrypt ( ECC_Key & key , const Octet_String & plaintext , Octet_String & cyphertext ) ;

    void ECC_B_233_decrypt ( ECC_Key & key , const Octet_String & cyphertext , Octet_String & plaintext ) ;


    void ECC_P_256_encrypt ( ECC_Key & key , const Octet_String & plaintext , Octet_String & cyphertext ) ;

    void ECC_P_256_decrypt ( ECC_Key & key , const Octet_String & cyphertext , Octet_String & plaintext ) ;

          
    void ECC_K_283_encrypt ( ECC_Key & key , const Octet_String & plaintext , Octet_String & cyphertext ) ;

    void ECC_K_283_decrypt ( ECC_Key & key , const Octet_String & cyphertext , Octet_String & plaintext ) ;


    void ECC_B_283_encrypt ( ECC_Key & key , const Octet_String & plaintext , Octet_String & cyphertext ) ;

    void ECC_B_283_decrypt ( ECC_Key & key , const Octet_String & cyphertext , Octet_String & plaintext ) ;


    void AES_ECB_128_encrypt ( Symmetric_Key & key , const Octet_String & plaintext , Octet_String & cyphertext ) ;

    void AES_ECB_128_decrypt ( Symmetric_Key & key , const Octet_String & cyphertext , Octet_String & plaintext ) ;

       
    void AES_CBC_128_encrypt ( Symmetric_Key & key , const Octet_String & plaintext , Octet_String & cyphertext ) ;

    void AES_CBC_128_decrypt ( Symmetric_Key & key , const Octet_String & cyphertext , Octet_String & plaintext ) ;

       
    void AES_ECB_192_encrypt ( Symmetric_Key & key , const Octet_String & plaintext , Octet_String & cyphertext ) ;

    void AES_ECB_192_decrypt ( Symmetric_Key & key , const Octet_String & cyphertext , Octet_String & plaintext ) ;

       
    void AES_CBC_192_encrypt ( Symmetric_Key & key , const Octet_String & plaintext , Octet_String & cyphertext ) ;

    void AES_CBC_192_decrypt ( Symmetric_Key & key , const Octet_String & cyphertext , Octet_String & plaintext ) ;

       
    void AES_ECB_256_encrypt ( Symmetric_Key & key , const Octet_String & plaintext , Octet_String & cyphertext ) ;

    void AES_ECB_256_decrypt ( Symmetric_Key & key , const Octet_String & cyphertext , Octet_String & plaintext ) ;

       
    void AES_CBC_256_encrypt ( Symmetric_Key & key , const Octet_String & plaintext , Octet_String & cyphertext ) ;

    void AES_CBC_256_decrypt ( Symmetric_Key & key , const Octet_String & cyphertext , Octet_String & plaintext ) ;


    RSA_Key_Pointer Generate_RSA_Key_Pair ( const unsigned int bits , const unsigned int exponent = 17 ) ;
    
    ECC_Key_Pointer Generate_ECC_Key_Pair ( const Octet mechanism ) ;

    void encrypt_symmetric ( Algorithm_Identifier algorithm_identifier , 
                             Key_Reference key_reference , 
                             const Octet_String & plaintext , 
                             Octet_String & cyphertext )
      {
        Symmetric_Key & key = validate_algorithm_and_symmetric_key ( algorithm_identifier , key_reference ) ;
        
        switch ( Octet ( algorithm_identifier ) )
          {
            case Algorithm_Identifier::Triple_DES_ECB_128:
              Triple_DES_ECB_128_encrypt ( key , plaintext , cyphertext ) ;
              break ;

            case Algorithm_Identifier::Triple_DES_CBC_128:
              Triple_DES_CBC_128_encrypt ( key , plaintext , cyphertext ) ;
              break ;

            case Algorithm_Identifier::Triple_DES_ECB_192:
              Triple_DES_ECB_192_encrypt ( key , plaintext , cyphertext ) ;
              break ;
          
            case Algorithm_Identifier::Triple_DES_CBC_192:
              Triple_DES_CBC_192_encrypt ( key , plaintext , cyphertext ) ;
              break ;
          
            case Algorithm_Identifier::AES_ECB_128:
              AES_ECB_128_encrypt ( key , plaintext , cyphertext ) ;
              break ;
          
            case Algorithm_Identifier::AES_CBC_128:
              AES_CBC_128_encrypt ( key , plaintext , cyphertext ) ;
              break ;
          
            case Algorithm_Identifier::AES_ECB_192:
              AES_ECB_192_encrypt ( key , plaintext , cyphertext ) ;
              break ;
          
            case Algorithm_Identifier::AES_CBC_192:
              AES_CBC_192_encrypt ( key , plaintext , cyphertext ) ;
              break ;
          
            case Algorithm_Identifier::AES_ECB_256:
              AES_ECB_256_encrypt ( key , plaintext , cyphertext ) ;
              break ;
          
            case Algorithm_Identifier::AES_CBC_256:
              AES_CBC_256_encrypt ( key , plaintext , cyphertext ) ;
              break ;
          }
      } ;

    // performs a private key cipher on the plaintext
    void encrypt_RSA ( Algorithm_Identifier algorithm_identifier , 
                       Key_Reference key_reference , 
                       const Octet_String & plaintext , 
                       Octet_String & cyphertext )
      {
        RSA_Key & key = validate_algorithm_and_RSA_key ( algorithm_identifier , key_reference ) ;
      
        switch ( Octet ( algorithm_identifier ) )
        {
            case Algorithm_Identifier::RSA_1024:
              RSA_1024_encrypt ( key , plaintext , cyphertext ) ;
              break ;
            
            case Algorithm_Identifier::RSA_2048:
              RSA_2048_encrypt ( key , plaintext , cyphertext ) ;
              break ;

            case Algorithm_Identifier::RSA_3072:
              RSA_3072_encrypt ( key , plaintext , cyphertext ) ;
              break ;
        }
      } ;

    void encrypt_ECC ( Algorithm_Identifier algorithm_identifier , 
                       Key_Reference key_reference , 
                       const Octet_String & plaintext , 
                       Octet_String & cyphertext )
      {
        ECC_Key & key = validate_algorithm_and_ECC_key ( algorithm_identifier , key_reference ) ;
        
        switch ( Octet ( algorithm_identifier ) )
        {
            case Algorithm_Identifier::ECC_P_224:
                ECC_P_224_encrypt ( key , plaintext , cyphertext ) ;
                break;

            case Algorithm_Identifier::ECC_K_233:
                ECC_K_233_encrypt ( key , plaintext , cyphertext ) ;
                break;

            case Algorithm_Identifier::ECC_B_233:
                ECC_B_233_encrypt ( key , plaintext , cyphertext ) ;
                break;

            case Algorithm_Identifier::ECC_P_256:
                ECC_P_256_encrypt ( key , plaintext , cyphertext ) ;
                break;

            case Algorithm_Identifier::ECC_K_283:
                ECC_K_283_encrypt ( key , plaintext , cyphertext ) ;
                break;

            case Algorithm_Identifier::ECC_B_283:
                ECC_B_283_encrypt ( key , plaintext , cyphertext ) ;
                break;

            default: break;
                throw RAPDU_Packet::SW_FUNCTION_NOT_SUPPORTED ;
        }
      } ;

    void encrypt ( const Octet_String & plaintext , Octet_String & cyphertext )
      {
        Algorithm_Identifier the_algorithm_identifier =
          the_security_environment . the_algorithm_identifier ;

        if ( the_algorithm_identifier . is_RSA_algorithm ( ) )
        {
          encrypt_RSA ( the_algorithm_identifier , 
                        the_security_environment . the_key_reference , 
                        plaintext , 
                        cyphertext ) ;
        }
        else if ( the_algorithm_identifier . is_ECC_algorithm ( ) )
        {
          encrypt_ECC ( the_algorithm_identifier , 
                        the_security_environment . the_key_reference , 
                        plaintext , 
                        cyphertext ) ;
        }
        else
        {
          encrypt_symmetric ( the_algorithm_identifier , 
                              the_security_environment . the_key_reference , 
                              plaintext , 
                              cyphertext ) ;
        }
      } ;


    void decrypt_symmetric ( Algorithm_Identifier algorithm_identifier , 
                             Key_Reference key_reference , 
                             const Octet_String & cyphertext , 
                             Octet_String & plaintext )
      {
        Symmetric_Key & key = validate_algorithm_and_symmetric_key ( algorithm_identifier , key_reference ) ;
        
        switch ( Octet ( algorithm_identifier ) )
          {
            case Algorithm_Identifier::Triple_DES_ECB_128:
              Triple_DES_ECB_128_decrypt ( key , cyphertext , plaintext ) ;
              break ;

            case Algorithm_Identifier::Triple_DES_CBC_128:
              Triple_DES_CBC_128_decrypt ( key , cyphertext , plaintext ) ;
              break ;

            case Algorithm_Identifier::Triple_DES_ECB_192:
              Triple_DES_ECB_192_decrypt ( key , cyphertext , plaintext ) ;
              break ;
          
            case Algorithm_Identifier::Triple_DES_CBC_192:
              Triple_DES_CBC_192_decrypt ( key , cyphertext , plaintext ) ;
              break ;
          
            case Algorithm_Identifier::AES_ECB_128:
              AES_ECB_128_decrypt ( key , cyphertext , plaintext ) ;
              break ;
          
            case Algorithm_Identifier::AES_CBC_128:
              AES_CBC_128_decrypt ( key , cyphertext , plaintext ) ;
              break ;
          
            case Algorithm_Identifier::AES_ECB_192:
              AES_ECB_192_decrypt ( key , cyphertext , plaintext ) ;
              break ;
          
            case Algorithm_Identifier::AES_CBC_192:
              AES_CBC_192_decrypt ( key , cyphertext , plaintext ) ;
              break ;
          
            case Algorithm_Identifier::AES_ECB_256:
              AES_ECB_256_decrypt ( key , cyphertext , plaintext ) ;
              break ;
          
            case Algorithm_Identifier::AES_CBC_256:
              AES_CBC_256_decrypt ( key , cyphertext , plaintext ) ;
              break ;
          }
      } ;

    // performs a public key cipher on the cyphertext
    void decrypt_RSA ( Algorithm_Identifier algorithm_identifier , 
                       Key_Reference key_reference , 
                       const Octet_String & cyphertext , 
                       Octet_String & plaintext )
      {
        RSA_Key & key = validate_algorithm_and_RSA_key ( algorithm_identifier , key_reference ) ;
        
        switch ( Octet ( algorithm_identifier ) )
          {
            case Algorithm_Identifier::RSA_1024:
              RSA_1024_decrypt ( key , cyphertext , plaintext ) ;
              break ;
            
            case Algorithm_Identifier::RSA_2048:
              RSA_2048_decrypt ( key , cyphertext , plaintext ) ;
              break ;
          }
      } ;

    void decrypt_ECC ( Algorithm_Identifier algorithm_identifier , 
                       Key_Reference key_reference , 
                       const Octet_String & cyphertext , 
                       Octet_String & plaintext )
      {
        ECC_Key & key = validate_algorithm_and_ECC_key ( algorithm_identifier , key_reference ) ;
        
        switch ( Octet ( algorithm_identifier ) )
        {
        case Algorithm_Identifier::ECC_P_224:
            break;
        case Algorithm_Identifier::ECC_K_233:
            break;
        case Algorithm_Identifier::ECC_B_233:
            break;
        case Algorithm_Identifier::ECC_P_256:
            break;
        case Algorithm_Identifier::ECC_K_283:
            break;
        case Algorithm_Identifier::ECC_B_283:
            break;
        default: break;
            throw RAPDU_Packet::SW_FUNCTION_NOT_SUPPORTED ;
        }
      } ;

    void decrypt ( const Octet_String & cyphertext , Octet_String & plaintext )
      {
        Algorithm_Identifier the_algorithm_identifier =
          the_security_environment . the_algorithm_identifier ;

        if ( the_algorithm_identifier . is_RSA_algorithm ( ) )
        {
          decrypt_RSA ( the_algorithm_identifier , 
                        the_security_environment . the_key_reference , 
                        cyphertext , 
                        plaintext ) ;
        }
        else if ( the_algorithm_identifier . is_ECC_algorithm ( ) )
        {
          decrypt_ECC ( the_algorithm_identifier , 
                        the_security_environment . the_key_reference , 
                        cyphertext , 
                        plaintext ) ;
        }
        else
        {
          decrypt_symmetric ( the_algorithm_identifier , 
                              the_security_environment . the_key_reference , 
                              cyphertext , 
                              plaintext ) ;
        }
    } ;


    bool authorized ( const Security_Attribute_Pointer security_attribute_pointer , 
                      Octet operation ) const 
      {
        return 
          security_attribute_pointer -> authorized ( * this , operation )  ;
      } ;

    // Debugging
    void dump ( ) ;

    bool entropy_is_enabled ;

    Security_Environment the_security_environment ;

    Security_Status the_security_status ;

    Key_Store the_key_store ;

    Key_Reference the_default_symmetric_key_number ;

    Key_Reference the_default_RSA_key_number ;
    
    Key_Reference the_default_PIN_number ;
  
  } ;

typedef indexed_pointer < Security_System >  Security_System_Pointer ;


#endif  // !defined ( __SECURITY_SYSTEM_H_INCLUDED__)
