// pivd - Personal Identity Verification Card Application Reference Implementation Daemon
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.
//

#include <stdio.h>
#include <Command_Line.h>
#include <Messages.h>

#include <TLP_APDU_Client.h>

#include <Card_Manager.h>
#include <File_Manager.h>

#define STRINGIZE1(a) #a
#define STRINGIZE(a) STRINGIZE1(a)
#define DEFAULT_APDUPORT_STRING STRINGIZE(TLP_APDU_Client::DEFAULT_APDU_PORT)

#define DEFAULT_PERSISTENT_STORAGE_SIZE 32768 //  8192
#define DEFAULT_PERSISTENT_STORAGE_SIZE_STRING STRINGIZE(DEFAULT_PERSISTENT_STORAGE_SIZE)

// Default settings for verbosity for Debug and Release when no verbosity is given on the command line
#if defined(_DEBUG)
  static int verbosity = 0 ;
#else //  defined(_DEBUG)
  static int verbosity = 0 ;
#endif  //  defined(_DEBUG)

static int port = TLP_APDU_Client::DEFAULT_APDU_PORT ;

static Memory::size_t the_persistent_storage_size = 
  DEFAULT_PERSISTENT_STORAGE_SIZE ;

static char * the_persistent_storage_input_filename = 0 ;

static char * the_persistent_storage_output_filename = 0 ;

static bool contactless = FALSE ;

#define  OPTION_HELP                0
#define  OPTION_VERBOSE             1
#define  OPTION_PORT                2
#define  OPTION_PERSISTENT_SIZE     3
#define  OPTION_PERSISTENT_INPUT    4
#define  OPTION_PERSISTENT_OUTPUT   5
#define  OPTION_TEST_FILESYSTEM     6
#define  OPTION_CONTACTLESS         7

option command_line_options [ ] =
{
   {
      'h',                          // option_char
      "help",                       // option_name
      OPTION_HELP,                  // option_num
      NO_PARM,                      // char parameter
      "print this message"          // description
   }
   ,
   {
      'v',                          // option_char
      "verbose",                    // option_name
      OPTION_VERBOSE,               // option_num
      NO_PARM,                      // char parameter
      "turn on debugging output -- use repeatedly for increasing chattiness"    // description
   }
   ,
   {
      'p',                          // option_char
      "port",                       // option_name
      OPTION_PORT,                  // option_num
      'p',                          // char parameter
      "listen for connections on port p (default=xxxxxxxxxxxxxxxxxxxx)"    // description
   }
   ,
   {
      's',                          // option_char
      "size",                       // option_name
      OPTION_PERSISTENT_SIZE,       // option_num
      'n',                          // char parameter
      "make initial persistent storage n Octets "
      "(default=" DEFAULT_PERSISTENT_STORAGE_SIZE_STRING  ", ignore if persistent file read in)"    // description
   }
   ,
   {
      'i',                          // option_char
      "input",                      // option_name
      OPTION_PERSISTENT_INPUT,      // option_num
      'f',                          // char parameter
      "read persistent storage from file f (default is to initialize persistent storage)"    // description
   }
   ,
   {
      'o',                          // option_char
      "output",                     // option_name
      OPTION_PERSISTENT_OUTPUT,     // option_num
      'f',                          // char parameter
      "write persistent storage to file f (default is to not to write persistent storage)"    // description
   }
   ,
   {
      'c',                          // option_char
      "contactless",                // option_name
      OPTION_CONTACTLESS,           // option_num
      NO_PARM,                      // char parameter
      "run on a contactless interface"  // description
   }
} ;

const int number_command_line_options = sizeof(command_line_options) / sizeof(option) ;

const char * usage_head =
"pivd - PIV Card Reference Implementation Daemon\n"
"usage: pivd [options]\n"
"  options:"
;

const char * usage_tail = NULL ;

bool process_option ( const int option, int & argc, char * * & argv )
{
   switch ( option )
   {
      case OPTION_HELP:
         show_usage ( command_line_options , number_command_line_options , usage_head , usage_tail ) ;
         exit ( 0 ) ;

      case OPTION_VERBOSE:
         verbosity ++ ;
         return true ;

      case OPTION_PORT:
         ++ argv ;
         -- argc ;
         if ( argc <= 0 )
         {
            error_message ( "Port number missing" ) ;
            return true ;
         }
         port = atoi ( * argv ) ;
         return true ;

      case OPTION_PERSISTENT_SIZE:
         ++ argv ;
         -- argc ;
         if ( argc <= 0 )
         {
            error_message ( "Persistent storage size missing" ) ;
            return true ;
         }
         the_persistent_storage_size = atoi ( * argv ) ;
         return true ;

      case OPTION_PERSISTENT_INPUT:
         ++ argv ;
         -- argc ;
         if ( argc <= 0 )
         {
            error_message ( "Persistent storage input file name missing" ) ;
            return true ;
         }
         the_persistent_storage_input_filename = strdup ( * argv ) ;
         return true ;


      case OPTION_PERSISTENT_OUTPUT:
         ++ argv ;
         -- argc ;
         if ( argc <= 0 )
         {
            error_message ( "Persistent storage output file name missing" ) ;
            return true ;
         }
         the_persistent_storage_output_filename = strdup ( * argv ) ;
         return true ;

       case OPTION_CONTACTLESS:
         contactless = TRUE ;        
         return true ;

      default:
         return false ;
   }
}



int main ( int argc , char * argv [ ] )
  {
    logfile = fopen ( "pivd.log" , "w" ) ;

    char * APDUPORT_environment_variable_value = getenv("APDUPORT") ;

    port = ( NULL != APDUPORT_environment_variable_value )
            ? atoi(APDUPORT_environment_variable_value)  
            : TLP_APDU_Client::DEFAULT_APDU_PORT ;

    char port_description [ 64 ]  ;
    snprintf ( port_description , sizeof ( port_description ) , 
              "listen for connections on port p (default=%d)" ,
              port ) ;
    command_line_options [ OPTION_PORT ] . description = port_description ;

    if ( ! process_command_line ( argc , 
                                  argv ,
                                  command_line_options ,
                                  number_command_line_options ,
                                  usage_head ,
                                  usage_tail ,
                                  process_option ) )
        return ( 255 ) ;

    if (0 < argc )
      {
          error_message ( "pivd: too many command line arguments\n") ;
          show_usage (command_line_options ,
                      number_command_line_options ,
                      usage_head ,
                      usage_tail);
          exit(255);
      }

    int return_code = 0 ;
    {
      Card_Manager the_card_manager ; 

      TLP_APDU_Client the_listener ( the_card_manager ) ;

      bool persistent_storage_loaded = false ;

      if ( (char *)0 != the_persistent_storage_input_filename )
        {
          FILE * the_persistent_storage_input_file =
            fopen ( the_persistent_storage_input_filename , "rb" ) ;
          if ( (FILE *)0 != the_persistent_storage_input_file )
            {
              unsigned short checksum = 0 ;
              unsigned short file_checksum = 0 ;
              persistent_storage_loaded = 
                ( Memory::load ( the_persistent_storage_input_file , checksum )
                    && the_card_manager . load_persistent_state 
                        (the_persistent_storage_input_file , checksum ) 
                    && 1 == fread ( & file_checksum , 
                                    sizeof ( file_checksum ) , 
                                    1 , 
                                    the_persistent_storage_input_file )
                    && checksum == file_checksum ) ;
              if ( ! persistent_storage_loaded )
                error_message ( "Failed reading the persistent storage input file \"%s\"\n" ,
                                the_persistent_storage_input_filename ) ;
              fclose ( the_persistent_storage_input_file ) ;
              // Debugging
//              Memory::dump_the_current_object_table( );
//              Memory::dump_the_free_list( );
            }
          else
            error_message ( "Failed opening persistent storage input file \"%s\"\n" ,
                            the_persistent_storage_input_filename  ) ;
          if ( ! persistent_storage_loaded )
            goto cleanup ;
        }

      if ( ! persistent_storage_loaded )
        {
          if ( 0 == the_persistent_storage_size )
            {
              error_message ( "No persistent storage input file and persistent storage size zero.\n" ) ;
              return_code |= 1 << 1;
              goto cleanup ;
            }

          persistent_storage_loaded =
            Memory::initialize ( the_persistent_storage_size ) ;

          if ( ! persistent_storage_loaded  )
            {
              error_message ( "Unable to allocate new persistent storage.\n" ) ;
              return_code |= 1 << 2;
              goto cleanup ;
            }
        }



      the_listener . set_port ( port );

//      the_listener . set_verbosity ( verbosity ) ;
//      Octet_String::set_verbosity ( verbosity );
//      Security_System::set_verbosity ( verbosity );
//      File_System::set_verbosity( verbosity );

      Security_System::set_interface( contactless ) ;

      if ( 0 != the_listener . initialize ( verbosity ) )
        {
          return_code |= 1 << 3 ;
          goto cleanup ;
        }

      if ( 0 != the_listener . run ( ) )
        return_code |= 1 << 4 ;

      if ( (char *)0 != the_persistent_storage_output_filename )
        {
          FILE * the_persistent_storage_output_file =
            fopen ( the_persistent_storage_output_filename , "wb" ) ;
          if ( (FILE *)0 != the_persistent_storage_output_file )
            {
//              Memory::dump_the_current_object_table( );
//              Memory::dump_the_free_list( );
              unsigned short int checksum = 0 ;
              if ( ! ( Memory::store 
                          ( the_persistent_storage_output_file , checksum  )
                      && the_card_manager . store_persistent_state
                            ( the_persistent_storage_output_file , checksum  ) 
                      && 1 == fwrite ( & checksum , 
                                       sizeof ( checksum ) , 
                                       1 , 
                                       the_persistent_storage_output_file ) ) )
                {
                  error_message ( "Failed writing persistent storage output file \"%s\"\n",
                                  the_persistent_storage_output_filename ) ;
                  return_code |= 1 << 5 ;
                }
              fclose ( the_persistent_storage_output_file ) ;
            }
          else
            {
              error_message ( "Failed opening persistent storage output file \"%s\"\n" ,
                              the_persistent_storage_output_filename  ) ;
              return_code |= 1 << 5 ;
            }
        }


      if ( 0 != the_listener . finalize ( ) )
        return_code |= 1 << 6 ;

    }

cleanup :

    Memory::finalize ( ) ;

    fclose ( logfile ) ;

    return return_code ;
  }

  #if defined(__cplusplus)
extern "C"
#endif
void
show_usage(const option options[],
           int NOPTIONS,
           const char *usage_head,
           const char *usage_tail)
{
  if (usage_head)
    error_message ("%s\n\n", usage_head);

  for (int n = 0; n < NOPTIONS; n++)
    {
      error_message ("   -%c %c  --%-24s   ",
                     options[n].option_char,
                     (options[n].parameter ? options[n].parameter : ' '),
                     options[n].option_name);
      char description[160] = { '\0' };
      strncpy (description, options[n].description,sizeof(description)/sizeof(char)-1);

      char *description2 = strchr (description, '\n');
      if (NULL == description2)
        error_message ("%s\n", description);
      else
        {
          *(description2++) = '\0';
          error_message ("%s\n", description);
          char * description3;
          while (NULL != ( description3 = strchr (description2, '\n') ))
            {
              *(description3++) = '\0';
              error_message ("%50s%s\n", description2 - 1, description2);
              description2 = description3 ;
            }
          error_message ("%50s%s\n", description2 - 1, description2);
        }
    }

  if (usage_tail)
    error_message ("\n%s\n", usage_tail);
}


#define SHOW_USAGE(error) \
  do \
    { \
      if ( file_being_processed ) \
        error_message ( "Error processing argument from %s:\n", file_being_processed ) ; \
      else \
        error_message ( "Error processing command-line argument:\n" ) ; \
      error_message error ; \
      show_usage ( options , \
                   NOPTIONS , \
                   usage_head , \
                   usage_tail ) ; \
      return false ; \
    } \
  while ( 0 )


static bool 
process_command_line_recursively (int &argc, char **&argv,
                                  const option options[],
                                  int NOPTIONS,
                                  const char *usage_head,
                                  const char *usage_tail, 
                                  PROCESSFN process_option,
                                  const char * file_being_processed)
{

  const char *exe_file_name = argv[0];

  while (argc--, argv++, 0 < argc) 
    switch ( argv[0][0] )
      {
        case '-':
          {
            char *p = argv[0] + 1;
            char c = *(p++);
            if ('-' == c)           //  --option [parameter]...
              {
                for (int n = 0; n < NOPTIONS; n++)
                  if (0 == strcmp (p, options[n].option_name))
                    if (!process_option (options[n].option_num, argc, argv))
                      SHOW_USAGE(("error processing option \"--%s\"\n", p));
                    else
                      break;
                if (n == NOPTIONS)
                  SHOW_USAGE(("unrecognized processing option \"--%s\"\n", p));
              }
            else if (!c)            //  isolated '-'
                SHOW_USAGE(("illegal option \"-\"", p));
            else                    //  -singleletteroptions [parameters-in-order]... 
              do
                {
                  for (int n = 0; n < NOPTIONS; n++)
                    if (options[n].option_char == c)
                      {
                        if (!process_option (options[n].option_num, argc, argv))
                          SHOW_USAGE(("error processing option \"-%c\"\n", c));
                        else
                          break;
                      }
                  if ( NOPTIONS == n )
                    SHOW_USAGE(("unrecognized option \"-%c\"\n", c));
                }
              while (c = *(p++));
          }
          break ;

        default:
          return true ;
      }
      
  return true;
}

static bool
convert_options_file_to_arglist ( const char * options_file , int & argc , char ** argv ) 
{
  char arg_buffer[4096];
  char * p = arg_buffer;
  char c ;
  FILE * options ;
  bool in_an_option = false ;

  if ( NULL == ( options = fopen ( options_file, "r" ) ) )
    return false ;

  argc = 1 ;
  argv[0] = strdup ( options_file ) ; 

  while ( c = fgetc( options ) , ! feof( options ) )
    if ( isspace ( c ) )
      {
        if ( in_an_option )
          {
            *p = '\0' ;
            argv [ argc ++ ] = strdup ( arg_buffer ) ;
            p = arg_buffer ;
            in_an_option = false ;
          }
      }
    else
      {
        if ( '\\' == c && ! feof ( options ) )
          {
            c = fgetc ( options ) ;
            if ( ! ( '"' == c ) )
              *p++ = '\\' ;
          }
        *p++ = c;
        in_an_option = true ;
      }

  if ( in_an_option )
    {
      *p = '\0' ;
      argv [ argc ++ ] = strdup ( arg_buffer ) ;
    }
  
  return true ;
}


static void
free_arglist( int & argc , char * * * pargv)
{
  for ( ; 0 < argc ; --argc )
    free ( * ( ( * pargv ) ++ ) );
}



bool process_command_line (int &argc, char **&argv,
                           const option options[],
                           int NOPTIONS,
                           const char *usage_head,
                           const char *usage_tail, 
                           PROCESSFN process_option)
{
  return process_command_line_recursively ( argc , argv, 
                                            options , 
                                            NOPTIONS , 
                                            usage_head , 
                                            usage_tail , 
                                            process_option , 
                                            NULL ) ;
}

#if defined(_WIN32) && defined(__cplusplus)
extern "C" {
#endif  /* defined(_WIN32) && defined(__cplusplus) */

#include <string.h>
#include <hexify.h>


static char hexdigits[] = "0123456789ABCDEF";


BOOL
ishex (char c)
{
  return '\0' != c  &&  NULL != strchr (hexdigits, toupper (c));
}

int
hexvalue ( char c )
{
  const char *pdigit = strchr (hexdigits, toupper (c));
  return static_cast < int > 
           ( ( (char *) NULL == pdigit) ? -1 : pdigit - hexdigits ) ;
}

char
hexchar ( unsigned int i )
{
  return i < 16 ? hexdigits[i] : '?';
}

BOOL
isallhex (const char *p)
{
  for (; *p;)
    {
      if (!ishex (*p++))
        return false;
    }
  return true;
}


int
ahntol (const char *buf, const int nchars, unsigned long * pvalue)
{
  unsigned long val = 0;
  for (int n = 0; n < nchars; n++)
    {
      int h = hexvalue (buf[n]);
      if ( -1 == h )
        return n + 1;
      val = (val << 4) | h ;
    }
  * pvalue = val;
  return 0;
}


int
ahtol (const char *buf, unsigned long * pvalue)
{
  unsigned long val = 0;
  for (const char *p = buf; *p; p++)
    {
      int h = hexvalue (*p);
      if ( -1 == h )
        return static_cast < int > ( buf - p + 1 ) ;
      val = (val << 4) | h ;
    }
  * pvalue = val;
  return 0;
}


char *
ltoh ( unsigned long value , char * buf )
{
  char * p = buf + sizeof ( value ) ;
  for (int i = 0 ; i < sizeof ( value ) ; i ++ , value >>= 4 )
    * -- p  = hexdigits [ value & 0xF ] ;
  return buf ;
}


char *
lntoh ( unsigned long value , int n , char * buf )
{
  char * p = buf + n ;
  for (int i = 0 ; i < n ; i ++ , value >>= 4 )
    * -- p = hexdigits [ value & 0xF ] ;
  return buf ;
}


char *
bntoh ( unsigned char * b , int n , char * buf )
{
  char * p = buf ;
  for (int i = 0 ; i < n ; i ++ , b ++  )
    {
      * p ++ = hexdigits [ ( * b ) >> 4  ] ;
      * p ++ = hexdigits [ ( * b ) & 0xF ] ;
    }
  return buf ;
}


#if defined(_WIN32) && defined(__cplusplus)
} /* extern "C" */
#endif  /* defined(_WIN32) && defined(__cplusplus) */

