// Octet_String - ( start , stop ) substrings of Octet storage
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#if !defined(__OCTET_STRING_H_INCLUDED__)
#define __OCTET_STRING_H_INCLUDED__


#include <Storage.h>
#include <iostream>

#include <Messages.h>

#if defined(_MSC_VER )
  // It's OK that the compiler can't automatically generate default constructors or
  // assignment operators for classes with const data members
  // We're shooting for immutability here
  #pragma warning ( disable : 4251 4510 4512 4610 )
  #if defined(OCTET_STRING_EXPORTS)
    #define OCTET_STRING_DLL _declspec(dllexport)
  #else //  defined ( OCTET_STRING_EXPORTS )
    #define OCTET_STRING_DLL _declspec(dllimport)
  #endif  // defined ( OCTET_STRING_EXPORTS )
#else
   #define OCTET_STRING_DLL
#endif

class OCTET_STRING_DLL Octet_String;

class OCTET_STRING_DLL Octet_String_Index_Exception
{
public:
  Octet_String_Index_Exception ( const Octet_String & octet_string , Memory::offset_t index ) 
      : the_octet_string ( octet_string ) , the_index ( index )
   {};
   const Octet_String & the_octet_string ;
   const Memory::offset_t the_index ;
};

class OCTET_STRING_DLL Octet_String :  public Object
{
public:

   // Create a NULL Octet_String
   Octet_String (  )
     : the_storage ( Octets::Pointer ( )  )
   {
     if ( 0 < get_verbosity ( ) )
       debug_message ( "Octet_String::CreateNull ( ) -- this=0x%08lX\n" ,  
                     reinterpret_cast < std::size_t > ( this ) ) ;
   };


   // Create a new Octet_String initialized to zero
   Octet_String ( size_t size )
      : the_start   ( 0      ) ,
        the_stop    ( size ) ,
        the_storage ( Octets::alloc ( size ) )
   {
     if ( 0 < get_verbosity ( ) )
       debug_message ( "Octet_String::CreatZeroized ( size=%d) -- this=0x%08lX the_storage=0x%08lX\n" ,  
                     size , 
                     reinterpret_cast < std::size_t > ( this ) , 
                     reinterpret_cast < std::size_t > ( & the_storage ) ) ;
   };


   // Create a new Octet_String based on a const byte array
   Octet_String ( const Octet * octets , size_t size )
      : the_start   ( 0    ) ,
        the_stop    ( size ) ,
        the_storage ( Octets::dup ( octets, size ) )
   {
     if ( 0 < get_verbosity ( ) ) 
       debug_message ( "Octet_String::CreateConstByteSize ( Octet *=0x%08lX, size=%d) -- this=0x%08lX the_storage=0x%08lX\n" , 
                     reinterpret_cast < std::size_t > ( octets ) , 
                     size ,
                     reinterpret_cast < std::size_t > ( this ) , 
                     reinterpret_cast < std::size_t > ( & the_storage ) ) ;
   };

   // Create a new Octet_String based on a const byte array
   Octet_String ( Octets::Pointer & storage , offset_t start , offset_t stop )
      : the_start   ( start ) ,
        the_stop    ( stop  ) ,
        the_storage ( storage )
   {
     if ( 0 < get_verbosity ( ) ) 
       debug_message ( "Octet_String::CreateConstStartStop ( Octets::Pointer=0x%08lX, start=%d, stop=%d)\n" , 
                     reinterpret_cast < std::size_t > ( & * storage ) , 
                     start , stop) ;
   };

   // Create a new Octet_String based on some existing storage
   Octet_String ( Octets::Pointer & storage , offset_t start = 0 )
      : the_start   ( start ) ,
        the_stop    ( storage -> get_size ( ) ) ,
        the_storage ( storage )
   {
     if ( 0 < get_verbosity ( ) ) 
       debug_message ( "Octet_String::CreateExisting ( Octets::Pointer=0x%08lX, start=%d)\n" , 
                     reinterpret_cast < std::size_t > ( & * storage ) , 
                     start ) ;
   };

   // Check substring bounds
   // Might turn this off unless debugging ...
   // ... if you're sure, that is.
   void 
   check_substring_bounds ( const Octet_String & other ) 
   throw ( Octet_String_Index_Exception ) 
     {
      if ( the_start < other .the_start )
        {
          throw Octet_String_Index_Exception ( * this, the_start ) ;
        }
      if ( other . the_stop < the_stop )
        {
          throw Octet_String_Index_Exception ( * this, the_stop) ;
        }
      if ( the_stop + 1 < the_start )
        {
          throw Octet_String_Index_Exception ( * this, the_start) ;
        }

      if ( 5 < get_verbosity ( ) )
        debug_message ( "Octet_String::Check ( const Octet_String &=0x%08lX) -- this=0x%08lX the_storage=0x%08lX\n", 
                        reinterpret_cast < std::size_t > ( & other ) , 
                        reinterpret_cast < std::size_t > ( this ) , 
                        reinterpret_cast < std::size_t > ( & * the_storage ) 
                      ) ;
    } ;

  // Create a new Octet_String as a substring of another Octet_String
  Octet_String ( const Octet_String & other, offset_t start , offset_t stop )
  throw ( Octet_String_Index_Exception ) 
    : the_storage ( other . the_storage ) ,
      the_start ( start) ,
      the_stop  ( stop )
    { check_substring_bounds ( other ) ; } ;

   // Create a new Octet_String as a substring of another Octet_String
   Octet_String ( const Octet_String & other, offset_t start )
   throw ( Octet_String_Index_Exception ) 
     : the_storage ( other . the_storage ) ,
       the_start ( start) ,
       the_stop  ( other . the_stop )
     { check_substring_bounds ( other ) ; } ;

   // Create a new Octet_String as a substring of another Octet_String
  Octet_String ( const Octet_String & other )
  throw ( Octet_String_Index_Exception )
    : the_storage ( other . the_storage ) ,
      the_start ( other . the_start) ,
      the_stop  ( other . the_stop )
    {
      // Might omit this check if you believe the invariant that every
      // constructed Octet_String is valid.  However, if someone
      // is going around setting the_start or the_stop ...
      // ... better leave this in.
      check_substring_bounds ( other ) ; 
    } ;

  // Zeroize an Octets
  static Octets::Pointer
  zeroize ( Octets::Pointer dst , size_t length )
    {
      memset ( dst -> body( ) , 0 , length ) ;
      return dst ;
    } ;

  static int
  cmp ( const Octet_String & dst , const Octet_String & src , size_t length )
    throw ( Octet_String_Index_Exception , out_of_range )
    {
      return Octets::cmp( dst . the_storage , dst . the_start , src . the_storage , src . the_start , length ) ;
    } ;

  bool
  operator < ( const Octet_String & other ) const
    {
      if ( length ( ) < other . length ( ) )
        return  cmp ( * this , other , length ( ) ) <= 0 ;
      else
        return cmp ( * this , other , other . length ( ) ) < 0 ;
    } ;

  // Copy one Octet_String into another Octet_String
  static Octet_String &
  cpy ( Octet_String & dst , const Octet_String & src , size_t length )
    throw ( Octet_String_Index_Exception , out_of_range )
    {
      if ( dst . the_stop < dst . the_start + length )
        throw Octet_String_Index_Exception ( dst , dst . the_stop ) ;
      if ( src . the_stop < src . the_start + length )
        throw Octet_String_Index_Exception ( src , src . the_stop ) ;
      Octets::cpy ( dst . the_storage , dst . the_start ,
                    src . the_storage , src . the_start ,
                    length ) ;
      return dst ;
    } ;

  static Octet_String &
  cpy ( Octet_String & dst , const Octet_String & src  )
    throw ( Octet_String_Index_Exception , out_of_range )
    {
      return cpy ( dst , src , src . the_stop - src . the_start ) ;
    } ;

  static Octet_String &
  cpy ( Octet_String & dst , offset_t dst_offset  , 
        const Octet_String & src ,  offset_t src_offset , 
        size_t length )
    throw ( Octet_String_Index_Exception , out_of_range )
    {
      if ( dst . the_stop < dst . the_start + length )
        throw Octet_String_Index_Exception ( dst , dst . the_stop ) ;
      if ( src . the_stop < src . the_start + length )
        throw Octet_String_Index_Exception ( src , src . the_stop ) ;
      Octets::cpy ( dst . the_storage , dst . the_start + dst_offset ,
                    src . the_storage , src . the_start + src_offset ,
                    length ) ;
      return dst ;
    } ;

  // Copy an Octet * into an Octet_String
  static Octet_String &
  cpy ( Octet_String & dst , const Octet * src , size_t length )
    throw ( Octet_String_Index_Exception , out_of_range )
    {
      if ( dst . the_stop < dst . the_start + length )
        throw Octet_String_Index_Exception ( dst , dst . the_stop  ) ;
      Octets::cpy ( dst . the_storage , dst . the_start , src , length ) ;
      return dst ;
    } ;

  // Copy an Octet_String into an Octet *
  static Octet *
  cpy ( Octet * dst , size_t dst_length , 
        const Octet_String & src , size_t src_length )
    throw ( Octet_String_Index_Exception , out_of_range )
    {
      if ( src . the_stop < src . the_start + src_length )
        throw Octet_String_Index_Exception ( src , src . the_stop  ) ;
      if ( dst_length < src_length )
        src_length = dst_length ;
      memcpy ( dst , src . the_storage -> body ( ) + src . the_start , src_length ) ;
      return dst ;
    } ;

  // Copy an Octet_String into an Octet *
  static Octet *
  cpy ( Octet * dst , size_t dst_length , 
        const Octet_String & src )
    throw ( Octet_String_Index_Exception , out_of_range )
    { return cpy ( dst , dst_length , src , src . length ( ) ) ; } ;

  // Copy an Octet_String into an Octets
  // Note:  This respects 'the_start' and 'the_stop' of the source.
  static Octets::Pointer &
  cpy ( Octets::Pointer & dst , offset_t dst_offset , const Octet_String & src )
    throw ( Octet_String_Index_Exception , out_of_range )
    {
      Octets::cpy ( dst , dst_offset , src . the_storage , src . start ( ) , src . length ( ) ) ;
      return dst ;
    } ;

  // Copy an Octets into an Octet_String
  // Note:  This respects 'the_start' of the destination.
  static Octet_String &
  cpy ( Octet_String & dst , 
        Octets::Pointer src , 
        offset_t offset , 
        size_t length  ) 
    {
      Octets::cpy ( dst . the_storage , dst . the_start  , src , offset , length ) ;
      return dst ;
    } ;


  // Copy an Octet_String into an Octet_String with independent storage
  static Octet_String
  dup ( const Octet_String & orig )
    throw ( Octet_String_Index_Exception , out_of_range )
    {
      Octet_String the_dup ( orig . length ( ) );
      Octet_String::cpy ( the_dup , orig ) ;
      return the_dup ;
    } ;

  // Concatenate two Octet_Strings
  static Octet_String
  concat ( const Octet_String & first,  const Octet_String & second )
    throw ( Octet_String_Index_Exception , out_of_range )
    {
      Octet_String the_concat ( first . length ( ) + second . length ( ) );
      if( first . length ( ) != 0 )
        Octet_String::cpy ( the_concat , first ) ;
      if( second . length ( ) != 0 )
        Octet_String::cpy ( the_concat , first . length ( ) , second , 0 , second . length ( ) ) ;
      return the_concat ;
    } ;

  // Copy out the storage of an Octet_String 
  static Octets::Pointer
  get_storage ( const Octet_String & orig )
    throw ( Octet_String_Index_Exception , out_of_range )
    {
      Octet_String the_dup ( orig . length ( ) );
      Octet_String::cpy ( the_dup , orig ) ;
      return the_dup . the_storage ;
    } ;


  Octet_String substring ( offset_t start )
    { 
      return Octet_String ( * this , the_start + start ) ;
    } ;


  Octet_String substring ( offset_t start , offset_t stop )
    { 
      return Octet_String ( * this , the_start + start , the_start + stop ) ;
    } ;

  bool 
  operator == ( const Octet_String & other ) const
    {
       return ( ( the_stop - the_start) 
                 == ( other.the_stop - other.the_start))
              && ( (  the_start == the_stop
              || 0 == Octets::cmp ( the_storage, the_start,
                                  other.the_storage , other.the_start,
                                  the_stop - the_start) ) ) ;
    } ;


  const Octet 
  operator [ ] ( offset_t index) const // throw ( ... ) 
    {
      return get_octet ( index) ;
    } ;

  const size_t length ( ) { return the_stop - the_start ; } ;

  // These are a last resort!  If you need them, what you really need
  // is probably a method on some subclass.
  const Octet 
  get_octet ( offset_t index ) const  // throw ( ... ) 
    {
      return the_storage -> get ( index + the_start ) ;
    } ;

  Octet 
  put_octet ( offset_t index, Octet o )  // throw ( ... ) 
    {
      return the_storage -> put ( index + the_start , o ) ;
    } ;

  const unsigned short int 
  get_u16 ( offset_t index ) const // throw ( ... ) 
    {
      return get_octet ( index ) * 256 + get_octet ( index + 1 ) ;
    } ;

  unsigned short int 
  put_u16 ( offset_t index, unsigned short int u16 )  // throw ( ... ) 
    {
      put_octet ( index     , static_cast <Octet> ( u16 / 256 ) ) ;
      put_octet ( index + 1 , static_cast <Octet> ( u16 % 256 ) ) ;
      return u16 ;
    } ;

  const short int 
  get_i16 ( offset_t index  ) const // throw ( ... ) 
    {
      Octet high_octet = the_storage -> get ( index + the_start ) ;
      return ( high_octet <= 127 
               ? high_octet 
               : high_octet - 256 ) * 256 
             + the_storage -> get ( index + the_start + 1 ) ;
    } ;

  short int 
  put_i16 ( offset_t index, short int i16 )  // throw ( ... ) 
    {
      the_storage -> put ( index + the_start     , static_cast <Octet> ( i16 / 256 ) ) ;
      the_storage -> put ( index + the_start + 1 , static_cast <Octet> ( i16 % 256 ) ) ;
      return i16 ;
    } ;

  const unsigned int 
  get_u24 ( offset_t index ) const // throw ( ... ) 
    {
      return ( (  (    the_storage -> get ( index + the_start     ) ) * 256 
                 +  the_storage -> get ( index + the_start + 1 ) ) * 256 
               + the_storage -> get ( index + the_start + 2 ) ) ;
    } ;

  unsigned int 
  put_u24 ( offset_t index, unsigned short int u32 )  // throw ( ... ) 
    {
      the_storage -> put ( index + the_start + 0 , static_cast <Octet> ( (  ( u32 / 256 ) / 256 ) % 256 ) ) ;
      the_storage -> put ( index + the_start + 1 , static_cast <Octet> ( (  u32 / 256 ) % 256 ) ) ;
      the_storage -> put ( index + the_start + 2 , static_cast <Octet> ( u32 % 256 ) ) ;
      return u32 ;
    } ;

  const unsigned int 
  get_u32 ( offset_t index ) const // throw ( ... ) 
    {
      return ( (  (   the_storage -> get ( index + the_start     ) ) * 256 
                 + the_storage -> get ( index + the_start + 1 ) ) * 256 
               + the_storage -> get ( index + the_start + 2 ) ) * 256
             +  the_storage -> get ( index + the_start + 3 ) ;
    } ;

  unsigned int 
  put_u32 ( offset_t index, unsigned short int u32 )  // throw ( ... ) 
    {
      the_storage -> put ( index + the_start ,     static_cast <Octet> ( (  ( u32 / 256 ) / 256 ) / 256 ) ) ;
      the_storage -> put ( index + the_start + 1 , static_cast <Octet> ( (  ( u32 / 256 ) / 256 ) % 256 ) ) ;
      the_storage -> put ( index + the_start + 2 , static_cast <Octet> ( (  u32 / 256 ) % 256 ) ) ;
      the_storage -> put ( index + the_start + 3 , static_cast <Octet> ( u32 % 256 ) ) ;
      return u32 ;
    } ;

  const short int 
  get_i32 ( offset_t index ) const // throw ( ... ) 
    {
      Octet high_octet = the_storage -> get ( index + the_start ) ;
      return ( (  ( (  high_octet <= 127 
                       ? high_octet 
                       : high_octet - 256           ) ) * 256
                 + the_storage -> get ( index + the_start + 1 ) ) * 256 
               + the_storage -> get ( index + the_start + 2 ) ) * 256 
             + the_storage -> get ( index + the_start + 3 ) ;
    } ;

  short int 
  put_i32 ( offset_t index, short int i32 )  // throw ( ... ) 
    {
      the_storage -> put ( index + the_start ,     static_cast <Octet> ( (  ( i32 / 256 ) / 256 ) / 256 ) ) ;
      the_storage -> put ( index + the_start + 1 , static_cast <Octet> ( (  ( i32 / 256 ) / 256 ) % 256 ) ) ;
      the_storage -> put ( index + the_start + 2 , static_cast <Octet> ( (  i32 / 256 ) % 256 ) ) ;
      the_storage -> put ( index + the_start + 3 , static_cast <Octet> ( i32 % 256 ) ) ;
      return i32 ;
    } ;

   ~Octet_String ( )
    {
      if ( 5 < get_verbosity ( ) ) 
        debug_message ( "Octet_String::~Octet_String ( ) -- this=0x%08lX the_storage=0x%08lX\n", 
                      reinterpret_cast < std::size_t > ( this ), 
                      reinterpret_cast < std::size_t > ( & the_storage ) ) ;
    } ;

  offset_t 
  start ( void ) const { return the_start ; } ;
   
  offset_t
  stop  ( void ) const { return the_stop  ; } ;

  size_t 
  length  ( void ) const { return the_stop - the_start  ; } ;

  string
  spell ( void ) const ;

  string
  spell_with_indentation ( string indentation ) const ;

  static int 
  get_verbosity ( void ) ;

  static int 
  set_verbosity ( int v ) ;

  // Don't use these if you can avoid it
  Octet * 
  data ( )
    { return & ( the_storage -> body ( ) ) [ the_start ]; } ;

  const Octet * 
  data ( ) const
    { return & ( the_storage -> body ( ) ) [ the_start ]; } ;

  bool
  store ( FILE * the_persistent_storage_output_file , unsigned short & checksum ) const ;

  bool 
  load ( FILE * the_persistent_storage_input_file , unsigned short & checksum ) ;

protected:

  Octets::Pointer the_storage ;
  
  offset_t  the_start ;
  offset_t  the_stop ;

  static int verbosity ;
} ;

typedef indexed_pointer < Octet_String > Octet_String_Pointer ;

#endif   // !defined(__OCTET_STRING_H_INCLUDED__)
